"""
First, a welcoming text, 'Welcome! We show some Symbols. Press SPACE to continue' is shown until
the participant presses the Spacebar. Then a square in purple is followed by a triangle in red
followed by a circle in green. All of these symbols are shown for 3000ms or until the participant
presses the key f or j. The correct key for the purple square and the red triangle is f, and
the correct key for the green circle is j.
"""

from sweetbean import Block, Experiment
from sweetbean.stimulus import Symbol, Text

# EVENT SEQUENCE

stim_1 = Text(
    text="Welcome! We show some Symbols. Press SPACE to continue",
    choices=[" "],
)

stim_2 = Symbol(
    duration=3000, symbol="square", color="#f0f", choices=["f", "j"], correct_key="f"
)

stim_3 = Symbol(
    duration=3000, symbol="triangle", color="red", choices=["f", "j"], correct_key="f"
)

stim_4 = Symbol(
    duration=3000, symbol="circle", color="green", choices=["f", "j"], correct_key="j"
)

event_sequence = [stim_1, stim_2, stim_3, stim_4]

# BLOCK DESIGN

block = Block(event_sequence)

# Create an image of the stimuli sequence of the block
block.to_image(path="symbols.png", data=None, zoom_factor=[1, 3, 3, 3])

# Create an HTML file of the experiment
experiment = Experiment([block])
experiment.to_html("symbols.html")



HTML_PREAMBLE = (
    "<!DOCTYPE html>\n"
    "<head>\n"
    "<title>My awesome experiment</title>"
    '<script src="https://unpkg.com/jspsych@7.3.1"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-html-keyboard-response@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-survey-text@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-survey-multi-choice@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-survey-likert@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-survey-likert@1.1.2"></script>\n'
    '<link href="https://unpkg.com/jspsych@7.3.1/css/jspsych.css" rel="stylesheet"'
    ' type="text/css"/>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-image-keyboard-response@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych/plugin-video-keyboard-response@1.1.2"></script>\n'
    '<script src="https://unpkg.com/@jspsych-contrib/plugin-rok@1.1.1"></script>\n'
    '<script src="https://unpkg.com/@jspsych-contrib/plugin-html-choice@1.0.0"></script>\n'
    '<script src="https://cdn.jsdelivr.net/npm/sweetbean@0.0.6/dist/runtime-script.js"></script>\n'
    '<link href="https://cdn.jsdelivr.net/npm/sweetbean@0.0.6/dist/style/main.css" '
    'rel="stylesheet" type="text/css"/>\n'
    '<link href="https://cdn.jsdelivr.net/npm/sweetbean@0.0.6/dist/style/bandit.css" '
    'rel="stylesheet" type="text/css"/>\n'
    "</head>\n"
    "<body></body>\n"
    "<script>\n"
)
HTML_APPENDIX = "</script>\n</html>"

JSPSYCH = {"jspsych": "7.3.1"}

DEPENDENCIES = {
    "jsPsychHtmlKeyboardResponse": {"@jspsych/plugin-html-keyboard-response": "1.1.2"},
    "jsPsychSurveyText": {"@jspsych/plugin-survey-text": "1.1.2"},
    "jsPsychSurveyMultiChoice": {"@jspsych/plugin-survey-multi-choice": "1.1.2"},
    "jsPsychSurveyLikert": {"@jspsych/plugin-survey-likert": "1.1.2"},
    "jsPsychRok": {"'@jspsych-contrib/plugin-rok": "1.1.1"},
    "jsPsychImageKeyboardResponse": {
        "@jspsych/plugin-image-keyboard-response": "1.1.2"
    },
    "jsPsychVideoKeyboardResponse": {
        "@jspsych/plugin-video-keyboard-response": "1.1.2"
    },
    "jsPsychHtmlChoice": {"@jspsych-contrib/plugin-html-choice": "1.0.0"},
}

AUTORA_PREAMBLE = (
    "import 'jspsych/css/jspsych.css'\nconst main = async (id, condition) => {\n"
    "const jsPsych = initJsPsych()\n"
)

AUTORA_APPENDIX = (
    "await jsPsych.run(trials)\nconst observation = jsPsych.data.get()\n"
    "return await observation\n}\nexport default main\n"
)


def FUNCTION_PREAMBLE(is_async):
    async_string = ""
    if is_async:
        async_string = "async "
    return f"{async_string}function runExperiment() " + (
        "{\n" "document.body.style.backgroundColor = 'black';\n"
    )


def FUNCTION_APPENDIX(is_async):
    async_string = ""
    if is_async:
        async_string = "await "
    return (
        f"{async_string}jsPsych.run(trials)\nconst observation = jsPsych.data.get()\n"
        + f"return {async_string}observation\n"
        + "}"
    )


def TEXT_APPENDIX(is_async):
    async_string = ""
    if is_async:
        async_string = "await "
    return f"{async_string}jsPsych.run(trials)\n"



from abc import ABC, abstractmethod

from sweetbean.util.parse import _fct_args_to_js, _fct_to_js, _var_to_js


class Variable(ABC):
    """
    A base class for variables.
    """

    def __init__(self, name):
        self.name = name

    @abstractmethod
    def to_js(self):
        pass


class TimelineVariable(Variable):
    """
    A timeline variable for jsPsych.
    """

    def to_js(self):
        return f"jsPsych.timelineVariable('{self.name}')"


class DataVariable(Variable):
    """
    A data variable for jsPsych to access data of previous trials.
    """

    def __init__(self, name, window):
        """
        Arguments:
            name: the name of the variable
            window: the window of the data
        """
        super().__init__(name)
        self.name = f'data["bean_{name}"]'
        self.raw_name = name
        self.window = window

    def to_js(self):
        if self.window == 0:
            return self.name
        return f"jsPsych.data.get().last({self.window})['trials'][0]['bean_{self.raw_name}']"


class FunctionVariable(Variable):
    """
    A variable that is the result of a function.
    Examples:
        >>> def add(a, b):
        ...     return a + b
        >>> FunctionVariable('a', add, [1, 2]).to_js()
        ((a,b) => {return a+b})(1, 2)

    """

    def __init__(self, name, fct, args):
        """
        Arguments:
            name: the name of the variable
            fct: the function
            args: the arguments of the function
        """
        super().__init__(name)
        self.fct = fct
        self.args = args

    def to_js(self):
        fct_declaration = _fct_to_js(self.fct)
        fct_input = _fct_args_to_js(self.args)
        return f"({fct_declaration}){fct_input}"


class CodeVariable(Variable):
    """
    A variable to access JavaScript code.
    """

    def __init__(self, name, value):
        """
        Arguments:
            name: the name of the variable
            value: the initial value of the variable
        """
        super().__init__(name)
        self.value = value

    def set(self):
        return f"let {self.name}={self.value};"

    def to_js(self):
        return self.name


class SharedVariable:
    """
    A variable that can be shared between different stimuli.
    """

    def __init__(self, name, value):
        """
        Arguments:
            name: the name of the variable
            value: the initial value of the variable
        """
        self.name = str(name)
        self.value = value

    def set(self):
        return f"let {self.name}={self.value};"

    def to_js(self):
        return self.name


class SideEffect:
    def __init__(self, set_variable, get_variable):
        """
        A side effect that can set variables.

        Arguments:
            set_variable: the variable to set (often a SharedVariable)
            get_variable: the variable to get (e.g, the variable the set variable should be set to)
        """
        self.set_variable = set_variable
        self.get_variable = get_variable

    def to_js(self):
        return f"{self.set_variable.name}={_var_to_js(self.get_variable)};"



import json

from sweetbean.stimulus.Stimulus import _KeyboardResponseStimulus


class Image(_KeyboardResponseStimulus):
    """
    shows an image
    """

    type = "jsPsychImageKeyboardResponse"

    def __init__(
        self,
        duration=None,
        stimulus="",
        choices=None,
        correct_key="",
        side_effects=None,
    ):
        """
        Arguments:
            duration: time in ms the stimulus is presented
            stimulus: the path to the image
            choices: the keys that will be recorded if pressed
            correct_key: the correct key to press
            side_effects: a dictionary of side effects
        """
        super().__init__(locals(), side_effects)

    def _add_special_param(self):
        pass

    def _set_before(self):
        pass

    def _get_prompt_l(self):
        try:
            with open("image_prompts.json") as f:
                image_prompts = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(
                "The file 'image_prompts.json' was not found. "
                "If you are using the Image stimulus, you need to provide a file "
                "'image_prompts.json' in the same directory as your experiment script. "
                "It should contain a dictionary with image paths as keys and prompts "
                "as values."
            )
        except json.JSONDecodeError:
            raise ValueError(
                "The file 'image_prompts.json' contains invalid JSON."
                "If you are using the Image stimulus, you need to provide a file "
                "'image_prompts.json' in the same directory as your experiment script. "
                "It should contain a dictionary with image paths as keys and prompts "
                "as values."
            )
        if not self.l_args["stimulus"] in image_prompts:
            raise ValueError(
                f"The image {self.l_args['stimulus']} is not in the image_prompts.json file."
                "If you are using the Image stimulus, you need to provide a file "
                "'image_prompts.json' in the same directory as your experiment script. "
                "It should contain a dictionary with image paths as keys and prompts "
                "as values."
            )

        return image_prompts[self.l_args["stimulus"]]



from sweetbean.stimulus.Choice import Bandit, HtmlChoice
from sweetbean.stimulus.HtmlKeyboardResponse import (
    Blank,
    Feedback,
    Fixation,
    Flanker,
    Stroop,
    Symbol,
    Text,
)
from sweetbean.stimulus.Image import Image
from sweetbean.stimulus.RO import RDP, ROK, RandomDotPatterns, RandomObjectKinematogram
from sweetbean.stimulus.Survey import LikertSurvey, MultiChoiceSurvey, TextSurvey
from sweetbean.stimulus.Video import Video



"""
A welcoming text the participant is shown until the participant presses the spacebar. The text says,
'Welcome! We show a survey. Press SPACE to continue'. This is followed by a survey with two
questions that the participant is supposed to answer on a likert scale. The first question is
'How are you feeling?' and the participant is supposed to rate their feeling on a scale from
-2 to 2 with 5 levels. The second question is 'Do you like this tool?' on a scale from 1 to 3.
Then, the participant is supposed to rate five animals on their cuteness from 1 to 5.
The animals are a bunny, a shark, a spider, a cat, and a dog.
"""

from sweetbean import Block, Experiment
from sweetbean.stimulus import LikertSurvey, Text

# EVENT SEQUENCE

stim_1 = Text(
    text="Welcome! We show a survey. Press SPACE to continue",
    choices=[" "],
)

stim_2 = LikertSurvey(
    questions=[
        {"prompt": "How are you feeling?", "labels": [-2, -1, 0, 1, 2]},
        {"prompt": "Do you like this tool?", "labels": [0, 1, 3]},
    ]
)

stim_3 = Text(
    text="Rate the following animals on a scale from 1-5 on their cuteness.",
    choices=[" "],
)

stim_4 = LikertSurvey.from_scale(
    prompts=["bunny", "shark", "spider", "cat", "dog"], scale=[1, 2, 3, 4, 5]
)

event_sequence = [stim_1, stim_2, stim_3, stim_4]

# BLOCK DESIGN

block = Block(event_sequence)
experiment = Experiment([block])
experiment.to_html("likert.html")



from sweetbean.stimulus.Stimulus import _KeyboardResponseStimulus


class Video(_KeyboardResponseStimulus):
    """
    shows a video
    """

    type = "jsPsychImageKeyboardResponse"

    def __init__(
        self,
        duration=0,
        src="",
        choices=None,
        correct_key="",
        side_effects=None,
    ):
        """
        Arguments:
            duration: time in ms the stimulus is presented
            src: the path to the image
            choices: the keys that will be recorded if pressed
            correct_key: the correct key to press
            side_effects: a dictionary of side effects
        """
        super().__init__(locals(), side_effects=side_effects)

    def _add_special_param(self):
        pass

    def _set_before(self):
        pass



"""
The text 'press a or b or wait' is shown in pink for 2000ms. The allowed responses are 'a' and 'b'.
"""

from sweetbean import Block, Experiment
from sweetbean.stimulus import Text

# EVENT SEQUENCE

text = Text(
    duration=2000, text="press a or b or wait", color="pink", choices=["a", "b"]
)


event_sequence = [text]

# BLOCK DESIGN

block = Block(event_sequence)
experiment = Experiment([block])

experiment.to_html("text.html")



from abc import ABC, abstractmethod
from typing import List, Union

from jinja2 import Template

from sweetbean.util.parse import to_js
from sweetbean.variable import (
    DataVariable,
    FunctionVariable,
    SharedVariable,
    TimelineVariable,
)


class _BaseStimulus(ABC):
    """
    A base class for stimuli
    """

    js = ""
    js_body = ""
    js_before = ""
    js_data = ""
    excludes: List[str] = []
    type = ""
    l_template: Union[str, None] = None
    l_args: dict = {}
    l_ses: dict = {}

    def __init__(self, args, side_effects=None):
        self.side_effects = side_effects
        if "self" in args:
            del args["self"]
        if "__class__" in args:
            del args["__class__"]
        if "side_effects" in args:
            del args["side_effects"]
        self.arg = args
        self.arg.update({"type": self.type})
        self.arg_js = {}
        if "duration" in self.arg:
            self.arg_js["trial_duration"] = self.arg["duration"]
        for key in self.arg:
            self.arg_js[key] = args[key]

    def return_shared_variables(self):
        shared_variables = {}
        for key in self.arg:

            def extract_shared_variables(value):
                if isinstance(value, SharedVariable):
                    shared_variables[value.name] = value
                elif isinstance(value, dict):
                    for v in value.values():
                        extract_shared_variables(v)
                elif isinstance(value, list):
                    for item in value:
                        extract_shared_variables(item)
                elif isinstance(value, FunctionVariable):
                    for arg in value.args:
                        extract_shared_variables(arg)

            extract_shared_variables(self.arg[key])
        return shared_variables

    def to_js(self):
        self.js = ""
        self.js_data = ""
        self.js_before = ""
        self.js_body = ""
        self._params_to_js()
        self.js = (
            f"{{{self.js_body}{self.js_before}on_finish:(data)=>{{{self.js_data}}}}}"
        )

    def to_js_for_image(self):
        self.js = ""
        self.js_data = ""
        self.js_before = ""
        self.js_body = ""
        self._params_to_js_from_prepared()
        self.js = (
            f"{{{self.js_body}{self.js_before}on_finish:(data)=>{{{self.js_data}}}}}"
        )

    def _params_to_js(self):
        self.js_body += f'type: {self.arg["type"]},'
        for key in self.arg_js:
            self._param_to_js(key, self.arg_js[key])
        for key in self.arg:
            if key not in self.arg_js:
                self._param_to_js_arg(key, self.arg[key])
        self._add_special_param()
        self._process_response()
        self._set_before()
        if self.side_effects:
            self._set_side_effects()

    def _params_to_js_from_prepared(self):
        self.js_body += f'type: {self.arg["type"]},'
        for key in self.l_args:
            self._param_to_js(key, self.l_args[key])
        self._add_special_param()
        self._process_response()
        self._set_before()
        if self.side_effects:
            self._set_side_effects()

    def _prepare_args_l(self, timeline_element, data, shared_variables, datum=None):
        if not datum:
            self.l_args = {}
            self.l_ses = {}
            for key, value in self.arg.items():
                key_ = key
                value_ = _parse_variable(
                    value, timeline_element, data, shared_variables
                )
                self.l_args[key_] = value_
        else:
            for key, value in datum.items():
                self.l_args[key] = value

    def _resolve_side_effects(self, timeline_element, data, shared_variables):
        if self.side_effects:
            for se in self.side_effects:
                get_variable = _parse_variable(
                    se.get_variable, timeline_element, data, shared_variables, se=True
                )
                self.l_ses[se.set_variable.name] = get_variable

    def process_l(self, prompts, get_input, multi_turn, datum=None):
        prompts.append(self._get_prompt_l())
        prompt_response = self._get_response_prompt_l()
        s_data = {}
        data = self.l_args.copy()
        if prompt_response:
            prompts[-1] += " " + prompt_response
            if multi_turn:
                _in_prompt = prompts[-1]
            else:
                _in_prompt = " ".join([p for p in prompts])
            if not datum:
                _r = get_input(_in_prompt)
                if isinstance(_r, str):
                    response = _r.upper()
                elif isinstance(_r, dict):
                    if "response" not in _r:
                        raise Exception(f"{_r} has an invalid response format")
                    response = _r["response"].upper()
                else:
                    raise Exception(f"{_r} has an invalid response format")
                # response = get_input(_in_prompt).upper()
            else:
                _r = datum["response"].upper()
                response = _r
            s_data = self._process_response_l(_r)
            prompts[-1] += f"{response}>>"
        data.update(s_data)
        return data, prompts

    def _get_prompt_l(self):
        if self.l_template is None:
            raise Exception("No template or function set for getting prompt")
        return Template(self.l_template).render(self.l_args)

    def _get_response_prompt_l(self):
        raise Exception("No template or function set for getting response prompt")

    def _process_response_l(self, response):
        if isinstance(response, dict):
            return response
        return {"response": response.upper()}

    def _param_to_js(self, key, param):
        body, data = _set_param_js(key, param)
        if key not in self.excludes and key != "type" and key != "duration":
            self.js_body += body
        self.js_data += data

    def _param_to_js_arg(self, key, param):
        _, data = _set_param_js(key, param)
        self.js_data += data

    def _set_side_effects(self):
        for se in self.side_effects:
            self.js_data += se.to_js()
            if not isinstance(se.set_variable, DataVariable):
                self.js_data += (
                    f'data["bean_{se.set_variable.name}"]={se.set_variable.name};'
                )

    @abstractmethod
    def _add_special_param(self):
        pass

    @abstractmethod
    def _process_response(self):
        pass

    @abstractmethod
    def _set_before(self):
        pass


class _KeyboardResponseStimulus(_BaseStimulus, ABC):
    """
    A base class for stimuli that require a correct key and choices
    """

    response_template = "You can press {{ choices }}. You press <<"

    response_key = "response"

    def _process_response(self):
        self.js_data += (
            f'data["bean_correct"]='
            f'data["bean_correct_key"]===data["{self.response_key}"];'
        )

    def _get_response_prompt_l(self):
        if not self.l_args["choices"]:
            return None
        return Template(self.response_template).render(
            {"choices": [c.upper() for c in self.l_args["choices"]]}
        )

    def _process_response_l(self, response):
        if not self.l_args["correct_key"]:
            return {"response": response.upper(), "correct": None}
        return {
            "response": response.upper(),
            "correct": response.upper() == self.l_args["correct_key"].upper(),
        }


def _set_param_js(key, param):
    body = _set_param_preamble(key)
    body += _set_set_variable(key, param)
    body += "return "
    body += _set_get_variable(key) + "},"
    data = _set_data_text(key, param)
    return body, data


def _set_param_preamble(param):
    return f"{param}:()=>{{"


def _set_set_variable(key, param):
    return f"let {key}={to_js(param)};"


def _set_get_variable(key):
    return f"{key}"


def _set_data_text(key, param):
    res = ""
    res += _set_set_variable(key, param)
    res += f'data["bean_{key}"]={_set_get_variable(key)};'
    return res


def _parse_variable(variable, timeline_element, data, shared_variables, se=False):
    if isinstance(variable, list):
        return [
            _parse_variable(a, timeline_element, data, shared_variables)
            for a in variable
        ]
    if isinstance(variable, dict):
        return {
            k: _parse_variable(v, timeline_element, data, shared_variables)
            for k, v in variable.items()
        }
    if isinstance(variable, tuple):
        return tuple(
            _parse_variable(a, timeline_element, data, shared_variables)
            for a in variable
        )
    if isinstance(variable, TimelineVariable):
        return timeline_element[variable.name]
    if isinstance(variable, DataVariable):
        if se:
            return data[-variable.window - 1][variable.raw_name]
        return data[-variable.window][variable.raw_name]
    if isinstance(variable, FunctionVariable):
        _args = [
            _parse_variable(a, timeline_element, data, shared_variables)
            for a in variable.args
        ]
        return variable.fct(*_args)
    if isinstance(variable, SharedVariable):
        return shared_variables[variable.name]

    return variable



import ast
import inspect
import os
import re
import subprocess
import tempfile
import textwrap


def to_js(var):
    return _var_to_js(var)


NON_LOCAL_INCLUDES = [
    "math",
    "random",
    "numpy",
    "pandas",
    "datetime",
    "time",
    "re",
    "os",
    "sys",
    "json",
    "csv",
]


def _fct_args_to_js(args):
    """
    Convert Python argument list to JavaScript inputs for a function.
    Examples:

    """
    res = "("
    for idx, arg in enumerate(args):
        res += str(_var_to_js(arg))
        if idx != len(args) - 1:
            res += ", "
    res += ")"
    return res


def _var_to_js(var):
    """
    Convert a python input value to a JavaScript input value.
    """
    if var is None:
        return "null"
    if isinstance(var, str):
        _var = var.replace("'", "\\'")
        return f"'{_var}'"
    # test if is sequence
    if isinstance(var, (list, tuple, set)):
        res = "["
        for idx, v in enumerate(var):
            res += str(_var_to_js(v))
            if idx != len(var) - 1:
                res += ","
        return f"{res}]"
    if isinstance(var, dict):
        res = "{"
        for k, v in var.items():
            res += f"'{k}': {_var_to_js(v)}"
            if k != list(var.keys())[-1]:
                res += ","
        return f"{res}}}"
    if isinstance(var, bool):
        return str(var).lower()
    if hasattr(var, "to_js"):
        return var.to_js()
    return var


def _fct_to_js(func):
    """
    Convert a Python function to JavaScript using Transcrypt.

    Examples:
    # >>> def add(a, b):
    # ...     return a + b
    # >>> _fct_to_js(add)
    # '(a,b) => {return a+b}'

    >>> a = lambda color: "f" if color == "red" else "j"
    >>> _fct_to_js(a)
    '(color) => {return __eq__(color,"red")?"f":"j"}'

    >>> b = lambda score, n: f"Score: {score/n}"
    >>> _fct_to_js(b)

    """
    global_vars = func.__globals__
    code = func.__code__

    # Find variables used in the function but not defined locally
    non_locals = []
    for varname in code.co_names:
        if varname in global_vars and varname not in NON_LOCAL_INCLUDES:
            non_locals.append(varname)
    if non_locals:
        raise ValueError(
            f"Function:\n"
            f"{inspect.getsource(func)}\n"
            f"contains non-local variables: {non_locals}.\n"
            f"Either the module is not supported "
            f"or the variable should be passed into "
            f"the FunctionVariable instead."
        )

    try:
        # Extract the source code of the function
        source_code = inspect.getsource(func)
        source_code = textwrap.dedent(source_code)
        source_code = replace_operators_with_functions(source_code)

        # Use a temporary directory to store the necessary files
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = os.path.join(temp_dir, "temp_script.py")

            # Write the function's source code to a temporary Python file
            with open(temp_path, "w") as temp_file:
                temp_file.write(source_code)

            if not os.path.exists(temp_path):
                raise FileNotFoundError(f"Temporary file not found at: {temp_path}")

            # Run Transcrypt to transpile the temporary file
            # subprocess.run(
            #     ["transcrypt", "-c", temp_path],
            #     check=True,
            #     stdout=subprocess.PIPE,
            #     stderr=subprocess.PIPE
            # )
            subprocess.run(
                ["transcrypt", "-b", temp_path],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            # Locate the JavaScript output in the `__javascript__` folder
            js_output_dir = os.path.join(temp_dir, "__target__")
            js_output_path = os.path.join(js_output_dir, "temp_script.js")

            # Check if the JavaScript file exists
            if not os.path.exists(js_output_path):
                raise FileNotFoundError(f"JavaScript file not found: {js_output_path}")

            # Read the full JavaScript code
            with open(js_output_path, "r") as js_file:
                full_js_code = js_file.read()

        return _extract_arrow_function(full_js_code)

    except Exception as e:
        return f"Error during conversion: {e}"


def _extract_arrow_function(js_code):
    js_code = _clean_function(js_code)

    match = re.search(r"function\((.*?)\)\{(.*?)\};", js_code, re.DOTALL)
    if not match:
        match = re.search(r"function\((.*?)\)\{(.*?)\},", js_code, re.DOTALL)
    if not match:
        match = re.search(r"function\((.*?)\)\{(.*?)\}", js_code, re.DOTALL)

    if match:
        # Convert to an arrow function
        params = match.group(1)
        body = match.group(2).strip()
        arrow_function = f"({params}) => {{{body}}}"
        arrow_function = _postprocess_format(arrow_function)
        arrow_function = _postprocess_functions(arrow_function)
        return arrow_function


def _clean_function(js_code):
    res = js_code.replace("__lambda__", "")
    res = res.replace("= function", "=function")
    res = res.replace("function (", "function(")
    res = res.replace(") =", ")=")
    return res


def _postprocess_format(js_code):
    """
    Replace `.format(...)` calls in the generated JavaScript
    with equivalent JavaScript template literals.

    Examples:
        >>> function = '(score,n) => {return"Score: {}".format(__truediv__(score,n))}'
        >>> _postprocess_format(function)
        '(score,n) => {return`Score: ${__truediv__(score,n)}`}'
    """
    import re

    # Match any string with `.format(...)`
    pattern = r'(["\'].*?\{.*?\}.*?["\'])\.format\((.*)\)'

    def replacer(match):
        """
        Replace `.format(...)` calls with JavaScript-compatible template literals.
        """
        # Extract the string template and the arguments
        template = match.group(1).strip()  # The string with placeholders (quoted)
        _inpt = match.group(2)
        # arguments = [
        #     arg.strip() for arg in match.group(2).split(",")
        # ]  # The replacement arguments
        arguments = _split_outer_commas(_inpt)
        # Remove the surrounding quotes and replace `{}` with `${...}`
        lst = template[-1]
        for i, arg in enumerate(template):
            j = i + 2
            candidate = template[-j]
            if candidate == lst:
                start = template[:-j] + "`"
                end = template[-j + 1 : -1] + "`"
                template = start + end
                break
        # template = template[1:-1]  # Strip the first and last character (quotes)
        for i, arg in enumerate(arguments):
            template = template.replace(
                "{}", f"${{{arg}}}", 1
            )  # Replace one placeholder at a time

        # Convert the modified string to a JavaScript template literal
        return f"{template}"

    # Replace all `.format(...)` calls in the JavaScript code
    js_code = re.sub(pattern, replacer, js_code)
    js_code = js_code.replace(".format()", "")
    return js_code


def _split_outer_commas(input_str):
    result = []
    current = []
    depth = 0  # Track depth of parentheses

    for char in input_str:
        if char == "," and depth == 0:  # Split only when at depth 0
            result.append("".join(current).strip())
            current = []
        else:
            if char == "(":
                depth += 1
            elif char == ")":
                depth -= 1
            current.append(char)

    # Append the last group
    if current:
        result.append("".join(current).strip())

    return result


def _postprocess_functions(js_code):
    """
    Replace known Python functions with their JavaScript equivalents.
    """
    # Define patterns and replacements
    replacements = {
        r"random\.randint\((.*?)\)": r"Math.floor(Math.random() * [\1][1]) - [\1][0] + 1",
        r"math.": r"Math.",
    }

    # Process replacements dynamically
    for py_function, js_function in replacements.items():
        js_code = re.sub(py_function, js_function, js_code)

    return js_code


OPERATOR_TO_FUNCTION = {
    ast.Add: "__add__",
    ast.Sub: "__sub__",
    ast.Mult: "__mul__",
    ast.Div: "__truediv__",
    ast.Mod: "__mod__",
    ast.Pow: "__pow__",
    ast.LShift: "__lshift__",
    ast.RShift: "__rshift__",
    ast.BitOr: "__or__",
    ast.BitXor: "__xor__",
    ast.BitAnd: "__and__",
    ast.FloorDiv: "__floordiv__",
    ast.MatMult: "__matmul__",
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__lt__",
    ast.LtE: "__le__",
    ast.Gt: "__gt__",
    ast.GtE: "__ge__",
}


class ReplaceOperatorsWithFunctions(ast.NodeTransformer):
    def visit_BinOp(self, node):
        self.generic_visit(node)
        operator_type = type(node.op)
        if operator_type in OPERATOR_TO_FUNCTION:
            func_name = OPERATOR_TO_FUNCTION[operator_type]
            new_node = ast.Call(
                func=ast.Name(id=func_name, ctx=ast.Load()),
                args=[node.left, node.right],
                keywords=[],
            )
            return ast.copy_location(new_node, node)
        else:
            return node

    def visit_UnaryOp(self, node):
        self.generic_visit(node)
        operator_type = type(node.op)
        if operator_type in OPERATOR_TO_FUNCTION:
            func_name = OPERATOR_TO_FUNCTION[operator_type]
            new_node = ast.Call(
                func=ast.Name(id=func_name, ctx=ast.Load()),
                args=[node.operand],
                keywords=[],
            )
            return ast.copy_location(new_node, node)
        else:
            return node

    def visit_Compare(self, node):
        self.generic_visit(node)
        if len(node.ops) == 1:
            operator_type = type(node.ops[0])
            if operator_type in OPERATOR_TO_FUNCTION:
                func_name = OPERATOR_TO_FUNCTION[operator_type]
                new_node = ast.Call(
                    func=ast.Name(id=func_name, ctx=ast.Load()),
                    args=[node.left, node.comparators[0]],
                    keywords=[],
                )
                return ast.copy_location(new_node, node)
        else:
            # Handle chained comparisons
            comparisons = []
            left = node.left
            for op, right in zip(node.ops, node.comparators):
                operator_type = type(op)
                if operator_type in OPERATOR_TO_FUNCTION:
                    func_name = OPERATOR_TO_FUNCTION[operator_type]
                    comparison = ast.Call(
                        func=ast.Name(id=func_name, ctx=ast.Load()),
                        args=[left, right],
                        keywords=[],
                    )
                    comparisons.append(comparison)
                    left = right
                else:
                    return node
            # Combine comparisons with 'and_'
            combined = comparisons[0]
            for comp in comparisons[1:]:
                combined = ast.Call(
                    func=ast.Name(id="and_", ctx=ast.Load()),
                    args=[combined, comp],
                    keywords=[],
                )
            return ast.copy_location(combined, node)
        return node

    def visit_BoolOp(self, node):
        self.generic_visit(node)
        operator_type = type(node.op)
        if operator_type in OPERATOR_TO_FUNCTION:
            func_name = OPERATOR_TO_FUNCTION[operator_type]
            # Since 'and' and 'or' are variable-length operators, we need to combine the values
            values = node.values
            combined = values[0]
            for value in values[1:]:
                combined = ast.Call(
                    func=ast.Name(id=func_name, ctx=ast.Load()),
                    args=[combined, value],
                    keywords=[],
                )
            return ast.copy_location(combined, node)
        else:
            return node

    def visit(self, node):
        return super().visit(node)


def replace_operators_with_functions(code):
    tree = ast.parse(code)
    transformer = ReplaceOperatorsWithFunctions()
    transformed_tree = transformer.visit(tree)
    ast.fix_missing_locations(transformed_tree)
    return ast.unparse(transformed_tree)



import importlib
import inspect
import os
import pkgutil
import sys

import sweetbean
from sweetbean import Block, Experiment
from sweetbean.stimulus.Stimulus import _BaseStimulus

# Define the excluded stimuli
excludes = [
    "SurveyStimulus",
    "_BaseStimulus",
    "_KeyboardResponseStimulus",
    "_Template_",
]


def test_compile():
    # Dynamically load all modules in the 'sweetbean.stimulus' package
    stimuli_package = sweetbean.stimulus

    # Dynamically load all modules in the 'sweetbean.stimulus' package
    for _, module_name, _ in pkgutil.iter_modules(stimuli_package.__path__):
        importlib.import_module(f"sweetbean.stimulus.{module_name}")

    # Gather all subclasses of Stimulus from the loaded modules
    stimuli_list = []
    for module in sys.modules.values():
        if module and module.__name__.startswith("sweetbean.stimulus"):
            for name, cls in inspect.getmembers(module, inspect.isclass):
                if issubclass(cls, _BaseStimulus) and cls.__name__ not in excludes:
                    stimuli_list.append(cls)

    # Debugging: Print the found stimuli
    print()
    print("Found stimuli:", stimuli_list)
    print()

    for stimulus in stimuli_list:
        # Log which stimulus is being tested
        print(f"Testing {stimulus.__name__}...")

        # Test each stimulus
        stimulus_instance = stimulus()  # Adjust if parameters are required
        trial_sequence = Block([stimulus_instance])
        experiment = Experiment([trial_sequence])
        experiment.to_html("basic.html")
        assert os.path.exists(
            "basic.html"
        ), f"{stimulus.__name__} failed to generate HTML."
        os.remove("basic.html")
        print(f"{stimulus.__name__} compiled successfully.")


if __name__ == "__main__":
    test_compile()



from sweetbean.stimulus.Stimulus import _BaseStimulus


class RandomObjectKinematogram(_BaseStimulus):
    """
    Show a random-object-kinematogram
    """

    type = "jsPsychRok"

    def __init__(
        self,
        duration=None,
        number_of_oobs=300,
        number_of_apertures=1,
        coherent_movement_direction=180,
        coherent_orientation=180,
        coherence_movement=100,
        coherence_orientation=100,
        movement_speed=10,
        aperture_position_left=50,
        aperture_position_top=50,
        oob_color="white",
        background_color="black",
        stimulus_type=0,
        choices=None,
        correct_key="",
        side_effects=None,
    ):
        """
        Arguments:
            duration: time in ms the stimulus is presented // trial_duration
            number_of_oobs: the number of oriented objects per set in the stimulus
            number_of_apertures: the number of kinematograms
            coherent_movement_direction: the direction of coherent motion in degrees
                (0 degre meaning right)
            coherent_orientation: the orientation of the objects in degree
                (0 degree meaning right)
            coherence_movement: the percentage of oriented objects moving in the coherent direction.
            coherence_orientation: the percentage of oriented objects moving in the coherent
                direction.
            movement_speed: the movement speed of the oobs in (percentage of aperture_width)/second
            aperture_position_left: position of midpoint of aperture in x direction in percentage
                of window width
            aperture_position_top: position of midpoint of aperture in y direction in percentage
                of window height
            oob_color: the color of the orientated objects
            background_color: the background color
            choices: the valid keys that the subject can press to indicate a response
            correct_key: the correct key to press
            side_effects: a dictionary of side effects
        """
        if choices is None:
            choices = []
        correct_choice = [correct_key]
        super().__init__(locals(), side_effects)

    def _add_special_param(self):
        pass

    def _process_response(self):
        self.js_data += 'data["bean_correct"]=data["correct"];'

    def _set_before(self):
        pass

    def process_l(self, prompts, get_input, multi_turn, datum=None):
        raise NotImplementedError


ROK = RandomObjectKinematogram


class RandomDotPatterns(ROK):
    def __init__(
        self,
        duration=None,
        number_of_oobs=300,
        number_of_apertures=2,
        coherent_orientation=0,
        coherence_orientation=0,
        aperture_position_left=50,
        aperture_position_top=50,
        oob_color="white",
        background_color="grey",
        stimulus_type=1,
        choices=None,
        correct_key="",
        side_effects=None,
    ):
        """
        Arguments:
            duration: time in ms the stimulus is presented // trial_duration
            number_of_oobs: the number of oriented objects per set in the stimulus
            number_of_apertures: the number of kinematograms
            coherent_orientation: the orientation of the objects in degree
                (0 degree meaning right)
            coherence_orientation: the percentage of oriented objects moving in the coherent
                direction.
            aperture_position_left: position of midpoint of aperture in x direction in percentage
                of window width
            aperture_position_top: position of midpoint of aperture in y direction in percentage
                of window height
            oob_color: the color of the orientated objects
            background_color: the background color
            choices: the valid keys that the subject can press to indicate a response
            correct_key: the correct key to press
            side_effects: a dictionary of side effects
        """
        if choices is None:
            choices = []

        super().__init__(
            duration=duration,
            number_of_oobs=number_of_oobs,
            number_of_apertures=number_of_apertures,
            coherent_movement_direction=0,
            coherent_orientation=coherent_orientation,
            coherence_movement=0,
            coherence_orientation=coherence_orientation,
            movement_speed=0,
            aperture_position_left=aperture_position_left,
            aperture_position_top=aperture_position_top,
            oob_color=oob_color,
            background_color=background_color,
            stimulus_type=stimulus_type,
            choices=choices,
            correct_key=correct_key,
            side_effects=side_effects,
        )


RDP = RandomDotPatterns
